@@ -27,26 +27,24 @@ functions scheduled by the Controller.
 import abc
 import collections
 import logging
-from monotonic import monotonic as now  # noqa
 import os
 import platform
+import queue
 import random
 import sys
 import threading
+import time
 import uuid
 
+from oslo_utils import eventletutils
 import proton
 import pyngus
-from six import iteritems
-from six import itervalues
-from six import moves
 
 from oslo_messaging._drivers.amqp1_driver.addressing import AddresserFactory
 from oslo_messaging._drivers.amqp1_driver.addressing import keyify
 from oslo_messaging._drivers.amqp1_driver.addressing import SERVICE_NOTIFY
 from oslo_messaging._drivers.amqp1_driver.addressing import SERVICE_RPC
 from oslo_messaging._drivers.amqp1_driver import eventloop
-from oslo_messaging._i18n import _LE, _LI, _LW
 from oslo_messaging import exceptions
 from oslo_messaging.target import Target
 from oslo_messaging import transport
@@ -81,7 +79,7 @@ class SubscribeTask(Task):
         self._subscriber_id = listener.id
         self._in_queue = listener.incoming
         self._service = SERVICE_NOTIFY if notifications else SERVICE_RPC
-        self._wakeup = threading.Event()
+        self._wakeup = eventletutils.Event()
 
     def wait(self):
         self._wakeup.wait()
@@ -108,21 +106,26 @@ class SendTask(Task):
         self.service = SERVICE_NOTIFY if notification else SERVICE_RPC
         self.timer = None
         self._retry = None if retry is None or retry < 0 else retry
-        self._wakeup = threading.Event()
+        self._wakeup = eventletutils.Event()
         self._error = None
+        self._sender = None
 
     def wait(self):
         self._wakeup.wait()
         return self._error
 
     def _execute(self, controller):
+        if self.deadline:
+            # time out the send
+            self.timer = controller.processor.alarm(self._on_timeout,
+                                                    self.deadline)
         controller.send(self)
 
     def _prepare(self, sender):
         """Called immediately before the message is handed off to the i/o
         system.  This implies that the sender link is up.
         """
-        pass
+        self._sender = sender
 
     def _on_ack(self, state, info):
         """If wait_for_ack is True, this is called by the eventloop thread when
@@ -143,10 +146,10 @@ class SendTask(Task):
         self._wakeup.set()
 
     def _on_timeout(self):
-        """Invoked by the eventloop when the send fails to complete before the
-        timeout is reached.
+        """Invoked by the eventloop when our timer expires
         """
         self.timer = None
+        self._sender and self._sender.cancel_send(self)
         msg = ("{name} message sent to {target} failed: timed"
                " out".format(name=self.name, target=self.target))
         LOG.warning("%s", msg)
@@ -172,6 +175,7 @@ class SendTask(Task):
         self._wakeup.set()
 
     def _cleanup(self):
+        self._sender = None
         if self.timer:
             self.timer.cancel()
             self.timer = None
@@ -190,9 +194,9 @@ class RPCCallTask(SendTask):
     """Performs an RPC Call.  Sends the request and waits for a response from
     the destination.
     """
-    def __init__(self, target, message, deadline, retry):
+    def __init__(self, target, message, deadline, retry, wait_for_ack):
         super(RPCCallTask, self).__init__("RPC Call", message, target,
-                                          deadline, retry, wait_for_ack=True)
+                                          deadline, retry, wait_for_ack)
         self._reply_link = None
         self._reply_msg = None
         self._msg_id = None
@@ -202,6 +206,7 @@ class RPCCallTask(SendTask):
         return error or self._reply_msg
 
     def _prepare(self, sender):
+        super(RPCCallTask, self)._prepare(sender)
         # reserve a message id for mapping the received response
         if self._msg_id:
             # already set so this is a re-transmit. To be safe cancel the old
@@ -214,7 +219,6 @@ class RPCCallTask(SendTask):
     def _on_reply(self, message):
         # called if/when the reply message arrives
         self._reply_msg = message
-        self._msg_id = None  # to prevent _cleanup() from cancelling it
         self._cleanup()
         self._wakeup.set()
 
@@ -226,10 +230,60 @@ class RPCCallTask(SendTask):
     def _cleanup(self):
         if self._msg_id:
             self._reply_link.cancel_response(self._msg_id)
+            self._msg_id = None
         self._reply_link = None
         super(RPCCallTask, self)._cleanup()
 
 
+class RPCMonitoredCallTask(RPCCallTask):
+    """An RPC call which expects a periodic heartbeat until the response is
+    received.  There are two timeouts:
+    deadline - overall hard timeout, implemented in RPCCallTask
+    call_monitor_timeout - keep alive timeout, reset when heartbeat arrives
+    """
+    def __init__(self, target, message, deadline, call_monitor_timeout,
+                 retry, wait_for_ack):
+        super(RPCMonitoredCallTask, self).__init__(target, message, deadline,
+                                                   retry, wait_for_ack)
+        assert call_monitor_timeout is not None  # nosec
+        self._monitor_timeout = call_monitor_timeout
+        self._monitor_timer = None
+        self._set_alarm = None
+
+    def _execute(self, controller):
+        self._set_alarm = controller.processor.defer
+        self._monitor_timer = self._set_alarm(self._call_timeout,
+                                              self._monitor_timeout)
+        super(RPCMonitoredCallTask, self)._execute(controller)
+
+    def _call_timeout(self):
+        # monitor_timeout expired
+        self._monitor_timer = None
+        self._sender and self._sender.cancel_send(self)
+        msg = ("{name} message sent to {target} failed: call monitor timed"
+               " out".format(name=self.name, target=self.target))
+        LOG.warning("%s", msg)
+        self._error = exceptions.MessagingTimeout(msg)
+        self._cleanup()
+        self._wakeup.set()
+
+    def _on_reply(self, message):
+        # if reply is null, then this is the call monitor heartbeat
+        if message.body is None:
+            self._monitor_timer.cancel()
+            self._monitor_timer = self._set_alarm(self._call_timeout,
+                                                  self._monitor_timeout)
+        else:
+            super(RPCMonitoredCallTask, self)._on_reply(message)
+
+    def _cleanup(self):
+        self._set_alarm = None
+        if self._monitor_timer:
+            self._monitor_timer.cancel()
+            self._monitor_timer = None
+        super(RPCMonitoredCallTask, self)._cleanup()
+
+
 class MessageDispositionTask(Task):
     """A task that updates the message disposition as accepted or released
     for a Server
@@ -250,7 +304,7 @@ class MessageDispositionTask(Task):
             self._disposition(self._released)
         except Exception as e:
             # there's really nothing we can do about a failed disposition.
-            LOG.exception(_LE("Message acknowledgment failed: %s"), e)
+            LOG.exception("Message acknowledgment failed: %s", e)
 
 
 class Sender(pyngus.SenderEventHandler):
@@ -329,23 +383,22 @@ class Sender(pyngus.SenderEventHandler):
     def send_message(self, send_task):
         """Send a message out the link.
         """
-        if send_task.deadline:
-            def timer_callback():
-                # may be in either list, or none
-                self._unacked.discard(send_task)
-                try:
-                    self._pending_sends.remove(send_task)
-                except ValueError:
-                    pass
-                send_task._on_timeout()
-            send_task.timer = self._scheduler.alarm(timer_callback,
-                                                    send_task.deadline)
-
         if not self._can_send or self._pending_sends:
             self._pending_sends.append(send_task)
         else:
             self._send(send_task)
 
+    def cancel_send(self, send_task):
+        """Attempts to cancel a send request.  It is possible that the send has
+        already completed, so this is best-effort.
+        """
+        # may be in either list, or none
+        self._unacked.discard(send_task)
+        try:
+            self._pending_sends.remove(send_task)
+        except ValueError:
+            pass
+
     # Pyngus callbacks:
 
     def sender_active(self, sender_link):
@@ -358,8 +411,8 @@ class Sender(pyngus.SenderEventHandler):
     def sender_remote_closed(self, sender_link, pn_condition):
         # The remote has initiated a close.  This could happen when the message
         # bus is shutting down, or it detected an error
-        LOG.warning(_LW("sender %(addr)s failed due to remote initiated close:"
-                        " condition=%(cond)s"),
+        LOG.warning("Sender %(addr)s failed due to remote initiated close:"
+                    " condition=%(cond)s",
                     {'addr': self._address, 'cond': pn_condition})
         self._link.close()
         # sender_closed() will be called once the link completes closing
@@ -369,7 +422,7 @@ class Sender(pyngus.SenderEventHandler):
 
     def sender_failed(self, sender_link, error):
         """Protocol error occurred."""
-        LOG.warning(_LW("sender %(addr)s failed error=%(error)s"),
+        LOG.warning("Sender %(addr)s failed error=%(error)s",
                     {'addr': self._address, 'error': error})
         self._handle_sender_closed(str(error))
 
@@ -431,8 +484,8 @@ class Sender(pyngus.SenderEventHandler):
                 self._unacked.discard(send_task)
                 if state == Sender._ACCEPTED:
                     send_task._on_ack(Sender._ACCEPTED, info)
-                elif (state == Sender._RELEASED
-                      or (state == Sender._MODIFIED and
+                elif (state == Sender._RELEASED or
+                      (state == Sender._MODIFIED and
                           # assuming delivery-failed means in-doubt:
                           not info.get("delivery-failed") and
                           not info.get("undeliverable-here"))):
@@ -537,10 +590,12 @@ class Replies(pyngus.ReceiverEventHandler):
 
     def prepare_for_response(self, request, callback):
         """Apply a unique message identifier to this request message. This will
-        be used to identify messages sent in reply.  The identifier is placed
-        in the 'id' field of the request message.  It is expected that the
-        identifier will appear in the 'correlation-id' field of the
+        be used to identify messages received in reply.  The identifier is
+        placed in the 'id' field of the request message.  It is expected that
+        the identifier will appear in the 'correlation-id' field of the
         corresponding response message.
+
+        When the caller is done receiving replies, it must call cancel_response
         """
         request.id = uuid.uuid4().hex
         # reply is placed on reply_queue
@@ -577,13 +632,13 @@ class Replies(pyngus.ReceiverEventHandler):
         receiver link has initiated closing the connection.
         """
         if pn_condition:
-            LOG.error(_LE("Reply subscription closed by peer: %s"),
+            LOG.error("Reply subscription closed by peer: %s",
                       pn_condition)
         receiver.close()
 
     def receiver_failed(self, receiver_link, error):
         """Protocol error occurred."""
-        LOG.error(_LE("Link to reply queue failed. error=%(error)s"),
+        LOG.error("Link to reply queue failed. error=%(error)s",
                   {"error": error})
         self._on_down()
 
@@ -597,12 +652,10 @@ class Replies(pyngus.ReceiverEventHandler):
         key = message.correlation_id
         try:
             self._correlation[key](message)
-            # cleanup (only need one response per request)
-            del self._correlation[key]
             receiver.message_accepted(handle)
         except KeyError:
-            LOG.warning(_LW("Can't find receiver for response msg id=%s, "
-                            "dropping!"), key)
+            LOG.warning("Can't find receiver for response msg id=%s, "
+                        "dropping!", key)
             receiver.message_modified(handle, True, True, None)
         # ensure we have enough credit
         if receiver.capacity <= self._capacity_low:
@@ -668,13 +721,13 @@ class Server(pyngus.ReceiverEventHandler):
                 "addr": receiver.source_address or receiver.target_address,
                 "err_msg": pn_condition
             }
-            LOG.error(_LE("Server subscription %(addr)s closed "
-                          "by peer: %(err_msg)s"), vals)
+            LOG.error("Server subscription %(addr)s closed "
+                      "by peer: %(err_msg)s", vals)
         receiver.close()
 
     def receiver_failed(self, receiver_link, error):
         """Protocol error occurred."""
-        LOG.error(_LE("Listener link queue failed. error=%(error)s"),
+        LOG.error("Listener link queue failed. error=%(error)s",
                   {"error": error})
         self.receiver_closed(receiver_link)
 
@@ -764,25 +817,23 @@ class NotificationServer(Server):
 
 class Hosts(object):
     """An order list of TransportHost addresses. Connection failover progresses
-    from one host to the next.  username, password , and realm come from the
-    configuration and are used only if no username/password/realm is present in
-    the URL.
+    from one host to the next.  The default realm comes from the configuration
+    and is only used if no realm is present in the URL.
     """
-    def __init__(self, entries=None, default_username=None,
-                 default_password=None,
-                 default_realm=None):
-        if entries:
-            self._entries = entries[:]
+    def __init__(self, url, default_realm=None):
+        self.virtual_host = url.virtual_host
+        if url.hosts:
+            self._entries = url.hosts[:]
         else:
             self._entries = [transport.TransportHost(hostname="localhost",
                                                      port=5672)]
         for entry in self._entries:
             entry.port = entry.port or 5672
-            entry.username = entry.username or default_username
-            entry.password = entry.password or default_password
+            entry.username = entry.username
+            entry.password = entry.password
             if default_realm and entry.username and '@' not in entry.username:
                 entry.username = entry.username + '@' + default_realm
-        self._current = random.randint(0, len(self._entries) - 1)
+        self._current = random.randint(0, len(self._entries) - 1)  # nosec
 
     @property
     def current(self):
@@ -797,7 +848,8 @@ class Hosts(object):
         return '<Hosts ' + str(self) + '>'
 
     def __str__(self):
-        return ", ".join(["%r" % th for th in self._entries])
+        r = ', vhost=%s' % self.virtual_host if self.virtual_host else ''
+        return ", ".join(["%r" % th for th in self._entries]) + r
 
 
 class Controller(pyngus.ConnectionEventHandler):
@@ -807,14 +859,14 @@ class Controller(pyngus.ConnectionEventHandler):
     work is done on the Eventloop thread, allowing the driver to run
     asynchronously from the messaging clients.
     """
-    def __init__(self, hosts, default_exchange, config):
+    def __init__(self, url, default_exchange, config):
         self.processor = None
         self._socket_connection = None
         self._node = platform.node() or "<UNKNOWN>"
         self._command = os.path.basename(sys.argv[0])
         self._pid = os.getpid()
         # queue of drivertask objects to execute on the eventloop thread
-        self._tasks = moves.queue.Queue(maxsize=500)
+        self._tasks = queue.Queue(maxsize=500)
         # limit the number of Task()'s to execute per call to _process_tasks().
         # This allows the eventloop main thread to return to servicing socket
         # I/O in a timely manner
@@ -837,14 +889,12 @@ class Controller(pyngus.ConnectionEventHandler):
         self.ssl_cert_file = config.oslo_messaging_amqp.ssl_cert_file
         self.ssl_key_file = config.oslo_messaging_amqp.ssl_key_file
         self.ssl_key_password = config.oslo_messaging_amqp.ssl_key_password
-        self.ssl_allow_insecure = \
-            config.oslo_messaging_amqp.allow_insecure_clients
+        self.ssl_verify_vhost = config.oslo_messaging_amqp.ssl_verify_vhost
+        self.pseudo_vhost = config.oslo_messaging_amqp.pseudo_vhost
         self.sasl_mechanisms = config.oslo_messaging_amqp.sasl_mechanisms
         self.sasl_config_dir = config.oslo_messaging_amqp.sasl_config_dir
         self.sasl_config_name = config.oslo_messaging_amqp.sasl_config_name
-        self.hosts = Hosts(hosts, config.oslo_messaging_amqp.username,
-                           config.oslo_messaging_amqp.password,
-                           config.oslo_messaging_amqp.sasl_default_realm)
+        self.hosts = Hosts(url, config.oslo_messaging_amqp.sasl_default_realm)
         self.conn_retry_interval = \
             config.oslo_messaging_amqp.connection_retry_interval
         self.conn_retry_backoff = \
@@ -903,13 +953,13 @@ class Controller(pyngus.ConnectionEventHandler):
 
     def shutdown(self, timeout=30):
         """Shutdown the messaging service."""
-        LOG.info(_LI("Shutting down the AMQP 1.0 connection"))
+        LOG.info("Shutting down the AMQP 1.0 connection")
         if self.processor:
             self.processor.wakeup(self._start_shutdown)
             LOG.debug("Waiting for eventloop to exit")
             self.processor.join(timeout)
             self._hard_reset("Shutting down")
-            for sender in itervalues(self._all_senders):
+            for sender in self._all_senders.values():
                 sender.destroy()
             self._all_senders.clear()
             self._servers.clear()
@@ -923,7 +973,7 @@ class Controller(pyngus.ConnectionEventHandler):
     # methods executed by Tasks created by the driver:
 
     def send(self, send_task):
-        if send_task.deadline and send_task.deadline <= now():
+        if send_task.deadline and send_task.deadline <= time.monotonic():
             send_task._on_timeout()
             return
         key = keyify(send_task.target, send_task.service)
@@ -972,22 +1022,46 @@ class Controller(pyngus.ConnectionEventHandler):
     def _do_connect(self):
         """Establish connection and reply subscription on processor thread."""
         host = self.hosts.current
-        conn_props = {'properties': {u'process': self._command,
-                                     u'pid': self._pid,
-                                     u'node': self._node},
-                      'hostname': host.hostname}
+        conn_props = {'properties': {'process': self._command,
+                                     'pid': self._pid,
+                                     'node': self._node}}
+        # only set hostname in the AMQP 1.0 Open performative if the message
+        # bus can interpret it as the virtual host.  We leave it unspecified
+        # since apparently noone can agree on how it should be used otherwise!
+        if self.hosts.virtual_host and not self.pseudo_vhost:
+            conn_props['hostname'] = self.hosts.virtual_host
         if self.idle_timeout:
             conn_props["idle-time-out"] = float(self.idle_timeout)
         if self.trace_protocol:
             conn_props["x-trace-protocol"] = self.trace_protocol
+
+        # SSL configuration
+        ssl_enabled = False
         if self.ssl:
+            ssl_enabled = True
             conn_props["x-ssl"] = self.ssl
         if self.ssl_ca_file:
             conn_props["x-ssl-ca-file"] = self.ssl_ca_file
+            ssl_enabled = True
         if self.ssl_cert_file:
+            ssl_enabled = True
             conn_props["x-ssl-identity"] = (self.ssl_cert_file,
                                             self.ssl_key_file,
                                             self.ssl_key_password)
+        if ssl_enabled:
+            # Set the identity of the remote server for SSL to use when
+            # verifying the received certificate.  Typically this is the DNS
+            # name used to set up the TCP connections.  However some servers
+            # may provide a certificate for the virtual host instead.  If that
+            # is the case we need to use the virtual hostname instead.
+            # Refer to SSL Server Name Indication (SNI) for the entire story:
+            # https://tools.ietf.org/html/rfc6066
+            if self.ssl_verify_vhost:
+                if self.hosts.virtual_host:
+                    conn_props['x-ssl-peer-name'] = self.hosts.virtual_host
+            else:
+                conn_props['x-ssl-peer-name'] = host.hostname
+
         # SASL configuration:
         if self.sasl_mechanisms:
             conn_props["x-sasl-mechs"] = self.sasl_mechanisms
@@ -1011,7 +1085,7 @@ class Controller(pyngus.ConnectionEventHandler):
             try:
                 self._tasks.get(False)._execute(self)
             except Exception as e:
-                LOG.exception(_LE("Error processing task: %s"), e)
+                LOG.exception("Error processing task: %s", e)
             count += 1
 
         # if we hit _max_task_batch, resume task processing later:
@@ -1052,10 +1126,13 @@ class Controller(pyngus.ConnectionEventHandler):
         point, we are ready to receive messages, so start all pending RPC
         requests.
         """
-        LOG.info(_LI("Messaging is active (%(hostname)s:%(port)s)"),
+        LOG.info("Messaging is active (%(hostname)s:%(port)s%(vhost)s)",
                  {'hostname': self.hosts.current.hostname,
-                  'port': self.hosts.current.port})
-        for sender in itervalues(self._all_senders):
+                  'port': self.hosts.current.port,
+                  'vhost': ("/" + self.hosts.virtual_host
+                            if self.hosts.virtual_host else "")})
+
+        for sender in self._all_senders.values():
             sender.attach(self._socket_connection.pyngus_conn,
                           self.reply_link, self.addresser)
 
@@ -1072,7 +1149,7 @@ class Controller(pyngus.ConnectionEventHandler):
 
     def socket_error(self, error):
         """Called by eventloop when a socket error occurs."""
-        LOG.error(_LE("Socket failure: %s"), error)
+        LOG.error("Socket failure: %s", error)
         self._handle_connection_loss(str(error))
 
     # Pyngus connection event callbacks (and their helpers), all invoked from
@@ -1099,9 +1176,11 @@ class Controller(pyngus.ConnectionEventHandler):
         # allocate an addresser based on the advertised properties of the
         # message bus
         props = connection.remote_properties or {}
-        self.addresser = self.addresser_factory(props)
-        for servers in itervalues(self._servers):
-            for server in itervalues(servers):
+        self.addresser = self.addresser_factory(props,
+                                                self.hosts.virtual_host
+                                                if self.pseudo_vhost else None)
+        for servers in self._servers.values():
+            for server in servers.values():
                 server.attach(self._socket_connection.pyngus_conn,
                               self.addresser)
         self.reply_link = Replies(self._socket_connection.pyngus_conn,
@@ -1131,7 +1210,7 @@ class Controller(pyngus.ConnectionEventHandler):
         # connection. Acknowledge the close, and try to reconnect/failover
         # later once the connection has closed (connection_closed is called).
         if reason:
-            LOG.info(_LI("Connection closed by peer: %s"), reason)
+            LOG.info("Connection closed by peer: %s", reason)
         self._detach_senders()
         self._detach_servers()
         self.reply_link.detach()
@@ -1144,8 +1223,8 @@ class Controller(pyngus.ConnectionEventHandler):
         """
         if outcome == proton.SASL.OK:
             return
-        LOG.error(_LE("AUTHENTICATION FAILURE: Cannot connect to "
-                      "%(hostname)s:%(port)s as user %(username)s"),
+        LOG.error("AUTHENTICATION FAILURE: Cannot connect to "
+                  "%(hostname)s:%(port)s as user %(username)s",
                   {'hostname': self.hosts.current.hostname,
                    'port': self.hosts.current.port,
                    'username': self.hosts.current.username})
@@ -1166,7 +1245,7 @@ class Controller(pyngus.ConnectionEventHandler):
             # service.  Try to re-establish the connection:
             if not self._reconnecting:
                 self._reconnecting = True
-                LOG.info(_LI("delaying reconnect attempt for %d seconds"),
+                LOG.info("Delaying reconnect attempt for %d seconds",
                          self._delay)
                 self.processor.defer(lambda: self._do_reconnect(reason),
                                      self._delay)
@@ -1184,9 +1263,9 @@ class Controller(pyngus.ConnectionEventHandler):
         if not self._closing:
             self._hard_reset(reason)
             host = self.hosts.next()
-            LOG.info(_LI("Reconnecting to: %(hostname)s:%(port)s"),
+            LOG.info("Reconnecting to: %(hostname)s:%(port)s",
                      {'hostname': host.hostname, 'port': host.port})
-            self._socket_connection.connect(host)
+            self.processor.wakeup(lambda: self._do_connect())
 
     def _hard_reset(self, reason):
         """Reset the controller to its pre-connection state"""
@@ -1198,7 +1277,7 @@ class Controller(pyngus.ConnectionEventHandler):
         del self._purged_senders[:]
         self._active_senders.clear()
         unused = []
-        for key, sender in iteritems(self._all_senders):
+        for key, sender in self._all_senders.items():
             # clean up any sender links that no longer have messages to send
             if sender.pending_messages == 0:
                 unused.append(key)
@@ -1208,8 +1287,8 @@ class Controller(pyngus.ConnectionEventHandler):
         for key in unused:
             self._all_senders[key].destroy(reason)
             del self._all_senders[key]
-        for servers in itervalues(self._servers):
-            for server in itervalues(servers):
+        for servers in self._servers.values():
+            for server in servers.values():
                 server.reset()
         if self.reply_link:
             self.reply_link.destroy()
@@ -1219,13 +1298,13 @@ class Controller(pyngus.ConnectionEventHandler):
 
     def _detach_senders(self):
         """Close all sender links"""
-        for sender in itervalues(self._all_senders):
+        for sender in self._all_senders.values():
             sender.detach()
 
     def _detach_servers(self):
         """Close all listener links"""
-        for servers in itervalues(self._servers):
-            for server in itervalues(servers):
+        for servers in self._servers.values():
+            for server in servers.values():
                 server.detach()
 
     def _purge_sender_links(self):
@@ -1251,5 +1330,5 @@ class Controller(pyngus.ConnectionEventHandler):
     @property
     def _active(self):
         # Is the connection up
-        return (self._socket_connection
-                and self._socket_connection.pyngus_conn.active)
+        return (self._socket_connection and
+                self._socket_connection.pyngus_conn.active)
