@@ -35,7 +35,6 @@ import kombu.messaging
 from oslo_config import cfg
 from oslo_log import log as logging
 from oslo_utils import eventletutils
-from oslo_utils import importutils
 
 import oslo_messaging
 from oslo_messaging._drivers import amqp as rpc_amqp
@@ -46,18 +45,6 @@ from oslo_messaging._drivers import pool
 from oslo_messaging import _utils
 from oslo_messaging import exceptions
 
-eventlet = importutils.try_import('eventlet')
-if eventlet and eventletutils.is_monkey_patched("thread"):
-    # Here we initialize module with the native python threading module
-    # if it was already monkey patched by eventlet/greenlet.
-    stdlib_threading = eventlet.patcher.original('threading')
-else:
-    # Manage the case where we run this driver in a non patched environment
-    # and where user even so configure the driver to run heartbeat through
-    # a python thread, if we don't do that when the heartbeat will start
-    # we will facing an issue by trying to override the threading module.
-    stdlib_threading = threading
-
 # NOTE(sileht): don't exist in py2 socket module
 TCP_USER_TIMEOUT = 18
 
@@ -465,14 +452,13 @@ class ConnectionLock(DummyConnectionLock):
 class Connection(object):
     """Connection object."""
 
-    def __init__(self, conf, url, purpose, retry=None):
+    def __init__(self, conf, url, purpose):
         # NOTE(viktors): Parse config options
         driver_conf = conf.oslo_messaging_rabbit
 
         self.interval_start = driver_conf.rabbit_retry_interval
         self.interval_stepping = driver_conf.rabbit_retry_backoff
         self.interval_max = driver_conf.rabbit_interval_max
-        self.max_retries = retry
 
         self.login_method = driver_conf.rabbit_login_method
         self.rabbit_ha_queues = driver_conf.rabbit_ha_queues
@@ -518,7 +504,7 @@ class Connection(object):
             # threading module with the native python threading module
             # if it was already monkey patched by eventlet/greenlet.
             global threading
-            threading = stdlib_threading
+            threading = _utils.stdlib_threading
 
         self.direct_mandatory_flag = driver_conf.direct_mandatory_flag
 
@@ -742,13 +728,7 @@ class Connection(object):
                       str(exc), interval)
 
         self._set_current_channel(None)
-        self.connection.ensure_connection(
-            errback=on_error,
-            max_retries=self.max_retries,
-            interval_start=self.interval_start or 1,
-            interval_step=self.interval_stepping,
-            interval_max=self.interval_max,
-        )
+        self.connection.ensure_connection(errback=on_error)
         self._set_current_channel(self.connection.channel())
         self.set_transport_socket_timeout()
 
